-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local collision = require('core/cameraModes/collision')

local function rotateVectorAroundZAxis(vector, angleDegrees)

    local angleRadians = math.rad(angleDegrees)

    local rotationMatrix = {
        {math.cos(angleRadians), -math.sin(angleRadians), 0},
        {math.sin(angleRadians), math.cos(angleRadians), 0},
        {0, 0, 1}
    }

    local rotatedVector = vec3(
            vector.x * rotationMatrix[1][1] + vector.y * rotationMatrix[1][2] + vector.z * rotationMatrix[1][3],
            vector.x * rotationMatrix[2][1] + vector.y * rotationMatrix[2][2] + vector.z * rotationMatrix[2][3],
            vector.x * rotationMatrix[3][1] + vector.y * rotationMatrix[3][2] + vector.z * rotationMatrix[3][3]
    )

    return rotatedVector
end

local C = {}
C.__index = C

function C:init()
  self.disabledByDefault = false
  self.camOffset = vec3(0, -5, 2)
  self.camResetted = 0
  self.lastVel = nil
  self.lastCamPos = vec3()
  self.lastDir = vec3()
  self.dirSmoothX = newTemporalSpring(10, 5)
  self.dirSmoothY = newTemporalSpring(20, 5)
  self.dirSmoothZ = newTemporalSpring(50, 5)
  self.dirSmoothV = newTemporalSpring(20, 5)
  self.dirSmoothA = newTemporalSpring(40, 1)
  self.dirSmoothD = newTemporalSpring(20, 5)
  self.smoothYaw = newTemporalSmoothingNonLinear(5, 5, 0)
  self.c_time = 0

  self.collision = collision()
  self.collision:init()

  self:onVehicleCameraConfigChanged()
  self:onSettingsChanged()
  self:reset()
end

function C:onVehicleCameraConfigChanged()
  self.fov = self.fov or 58
end

function C:onVehicleResetted()
  self:reset()
end

function C:onSettingsChanged()
  self:reset() --TODO is this really necessary?
end

function C:reset()
  self.camResetted = 0
  self.lastVel = nil
end

function C:update(data)
  data.res.collisionCompatible = true

  self.c_time = self.c_time + data.dt

  local camerarot = (MoveManager.yawRight - MoveManager.yawLeft) * -90
  camerarot = self.smoothYaw:get(camerarot, data.dt)

  local ref  = data.veh:getNodePosition(self.refNodes.ref)
  local left = data.veh:getNodePosition(self.refNodes.left)
  local back = data.veh:getNodePosition(self.refNodes.back)
  local center = vec3(); center:setLerp(back, left, 0.5);

  local dir = (ref - back)
  dir = dir:normalized()

  local quat_dir = quatFromDir(dir)
  local rev_dir = quat_dir:conjugated()

  if self.lastVel == nil then
    self.lastVel = dir
  end

  local dvel = data.vel - data.prevVel
  local accel = (dvel / data.dt)
  local l_accel = accel:rotated(rev_dir)
  local l_vel = data.vel:rotated(rev_dir)
  local vel_dir = data.vel:normalized()

  vel_dir:setLerp(self.lastVel, vel_dir, 0.06)
  if data.vel:length() > 1 then
    self.lastVel = vel_dir
  else
    vel_dir = self.lastVel
  end
  local dir_diff = vec3(); dir_diff:setLerp(dir, vel_dir, 0.75); dir_diff:normalized(); dir_diff = dir_diff:z0()
  if data.lookBack == true then
    camerarot = 180
  end
  dir_diff = rotateVectorAroundZAxis(dir_diff, camerarot)
  local vel_dir_quat = quatFromDir(dir_diff)

  local vel_s = clamp(math.exp((-100 / data.vel:length())+2), 0, 1)
  local shake = vel_s * math.sin(2 * math.pi * 10 * self.c_time)

  local camOffset = vec3(0, 0, 0)

  local x_offset = self.dirSmoothX:get(rescale(l_vel.x, -20, 20, 1, -1) + rescale(l_accel.x, -20, 20, -0.2, 0.2), data.dt)

  camOffset.x = x_offset
  camOffset.y = self.dirSmoothY:get(self.camOffset.y - (linearScale(math.abs(l_vel.y), 0, 50, 0, 1.5) + rescale(l_accel.y, -20, 20, -1, 2)), data.dt)
  camOffset.z = self.camOffset.z - (linearScale(math.abs(l_vel.y), 0, 50, 0, 0.6))


  camOffset.z = camOffset.z + shake * 0.05

  camOffset:setRotate(vel_dir_quat)


  local targetOffset = vec3(0, 5, 1)
  targetOffset.x = x_offset * 0.6

  targetOffset:setRotate(vel_dir_quat)

  local targetPos

  targetPos = data.pos + center + targetOffset

  -- application
  data.res.fov = linearScale(l_vel.y, 0, 100, self.fov, 90)
  data.res.pos = data.pos + center + camOffset
  data.res.rot = quatFromDir(targetPos - data.res.pos)
  if self.camResetted == 0 then
    self.dirSmoothZ:set(data.res.pos.z)
    self.camResetted = 1
  end

  local z_bounds = data.res.pos.z

  data.res.pos.z = clamp(self.dirSmoothZ:get(data.res.pos.z, data.dt), z_bounds - 1, z_bounds + 1)

  --debugDrawer:drawSphere((targetPos):toPoint3F(), 0.2, ColorF(0,0,1,1))
  --debugDrawer:drawLine((data.pos + center):toPoint3F(), (data.pos + center + dir):toPoint3F(), ColorF(0,1,0,1))
  --debugDrawer:drawLine((data.pos + center):toPoint3F(), (data.pos + center + data.vel):toPoint3F(), ColorF(0,1,1,1))
  --debugDrawer:drawLine((data.pos + center + data.vel):toPoint3F(), (data.pos + center + data.vel + accel):toPoint3F(), ColorF(1,1,0,1))
  --debugDrawer:setLastZTest(false)

  self.collision:update(data)
  return true
end

function C:setRefNodes(centerNodeID, leftNodeID, backNodeID)
  self.refNodes = self.refNodes or {}
  self.refNodes.ref = centerNodeID
  self.refNodes.left = leftNodeID
  self.refNodes.back = backNodeID
end

function C:mouseLocked(locked)
  if locked then return end
  self.relYaw = 0
  self.relPitch = 0
end

-- DO NOT CHANGE CLASS IMPLEMENTATION BELOW

return function(...)
  local o = ... or {}
  setmetatable(o, C)
  o:init()
  return o
end
